/*******************************************************************************
 * Copyright 2018 Frank Mnster
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/
package de.frankmuenster.jameica.qrcodehandler.bcdsepa;

import java.rmi.RemoteException;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import org.apache.commons.lang.StringUtils;

import de.frankmuenster.jameica.qrscanner.DefaultSepaTransaction;
import de.frankmuenster.jameica.qrscanner.Settings;
import de.frankmuenster.jameica.qrscanner.utils.IbanAndBicValidator;
import de.willuhn.jameica.gui.GUI;
import de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;

/**
 * Klasse f&uuml;r ein BCD Transfer Object. Das Object enth&auml;lt alle Felder, die im BCD-SCT Standard definiert sind
 * und ein Methode zum Erstellen einer Auslandsueberweisung in Hibiscus
 *
 * @author Frank Mnster
 */
@SuppressWarnings({ "PMD.GodClass", "nls" })
public class BcdDataTransferObject {

    /**
     *
     */
    private static final int HINT_MAX_LENGTH = 70;

    /**
     *
     */
    private static final String FUNCTION_SCT = "SCT"; //$NON-NLS-1$

    /**
     *
     */
    private static final int ZWECK_MAX_LENGTH = 140;

    /**
     * Die Servicekennung des SPEACredirTransfer Objects. Muss BCT lauten.
     */
    private String serviceKennung = "BCD"; //$NON-NLS-1$

    /**
     * Die Version des SPEACredirTransfer Objects.
     */
    private String version = StringUtils.EMPTY;

    /**
     * Die verwendete Kodierung des SPEACredirTransfer Objects. Dabei gilt: <br/>
     * 1: UTF-8<br/>
     * 2: ISO 8895-1<br/>
     * 3: ISO 8895-2<br/>
     * 4: ISO 8895-4<br/>
     * 5: ISO 8895-5<br/>
     * 6: ISO 8895-7<br/>
     * 7: ISO 8895-10 und<br/>
     * 8: ISO 8895-15
     */
    private String kodierung = StringUtils.EMPTY;

    /**
     * Die des SPEACredirTransfer Objects
     */
    private String funktion = BcdDataTransferObject.FUNCTION_SCT;

    /**
     * Die des SPEACredirTransfer Objects
     */
    private String bic = StringUtils.EMPTY;

    /**
     * Die des SPEACredirTransfer Objects
     */
    private String empfaenger = StringUtils.EMPTY;

    /**
     * Die des SPEACredirTransfer Objects
     */
    private String iban = StringUtils.EMPTY;

    /**
     * Die des SPEACredirTransfer Objects
     */
    private String betrag = StringUtils.EMPTY;

    /**
     * Die des SPEACredirTransfer Objects
     */
    private String geschaeftsCode = StringUtils.EMPTY;

    /**
     * Die des SPEACredirTransfer Objects
     */
    private String referenze = StringUtils.EMPTY;

    /**
     * Die des SPEACredirTransfer Objects
     */
    private String verwendungszweck = StringUtils.EMPTY;

    /**
     * Die des SPEACredirTransfer Objects
     */
    private String benutzerhinweis = StringUtils.EMPTY;

    /**
     * Erzeugt ein neues BcdDataTransferObject
     */
    public BcdDataTransferObject() {
        super();
    }

    /**
     * @return the serviceKennung
     */
    public String getServiceKennung() {
        return this.serviceKennung;
    }

    /**
     * @param serviceKennung
     *            the serviceKennung to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn die Servicekennung nicht der Definition entspricht. Es ist nur der Wert 'BCD'
     *             erlaubt.
     */
    public void setServiceKennung(final String serviceKennung) throws BcdDataTransferObjcetDefinitionException {
        if (serviceKennung == null) {
            throw new BcdDataTransferObjcetDefinitionException("Die Servicekennung darf nicht null sein.");
        }
        if (!serviceKennung.toUpperCase(Locale.getDefault()).contentEquals("BCD")) {
            throw new BcdDataTransferObjcetDefinitionException("Die Servicekennung muss 'BCD' lauten");
        }
        this.serviceKennung = serviceKennung.toUpperCase(Locale.getDefault());
    }

    /**
     * @return the version
     */
    public String getVersion() {
        return this.version;
    }

    /**
     * @param version
     *            the version to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn die Version null ist.
     */
    public void setVersion(final String version) throws BcdDataTransferObjcetDefinitionException {
        if (version == null) {
            throw new BcdDataTransferObjcetDefinitionException("Die Version darf nicht null sein.");
        }
        if (!"001".equals(version) && !"002".equals(version)) {
            throw new BcdDataTransferObjcetDefinitionException("Die Version kann nur 001 oder 002 sein.");
        }
        this.version = version;
    }

    /**
     * @return the kodierung
     */
    public String getKodierung() {
        return this.kodierung;
    }

    /**
     * @param kodierung
     *            the kodierung to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn die Kodierung nicht der Definition entspricht (Ziffer zwischen 0 und 8)
     */
    public void setKodierung(final String kodierung) throws BcdDataTransferObjcetDefinitionException {
        if (kodierung == null) {
            throw new BcdDataTransferObjcetDefinitionException("Die Kodierung darf nicht null sein.");
        }
        try {
            final Integer kod = Integer.decode(kodierung);
            if (kod >= 0 || kod <= 8) {
                this.kodierung = String.valueOf(kod);
                return;
            }
        } catch (final NumberFormatException e) {
            throw new BcdDataTransferObjcetDefinitionException(
                    "Die Kodierung " + kodierung + " ist fehlerhaft. Muss eine Ziffer zwischen 0 unn 8 sein", e);
        }
        throw new BcdDataTransferObjcetDefinitionException(
                "Die Kodierung " + kodierung + " ist fehlerhaft. Muss eine Ziffer zwischen 0 und 8 sein");
    }

    /**
     * @return the funktion
     */
    public String getFunktion() {
        return this.funktion;
    }

    /**
     * @param funktion
     *            the funktion to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn die Funktion nicht der Definition entspricht. Zur Zeit ist nur 'SCT' als
     *             Funktionstring erlaubt
     */
    public void setFunktion(final String funktion) throws BcdDataTransferObjcetDefinitionException {
        if (funktion == null) {
            throw new BcdDataTransferObjcetDefinitionException("Die Funktion darf nicht null sein.");
        }
        if (funktion.toUpperCase(Locale.getDefault()).contentEquals(BcdDataTransferObject.FUNCTION_SCT)) {
            this.funktion = BcdDataTransferObject.FUNCTION_SCT;
            return;
        }
        throw new BcdDataTransferObjcetDefinitionException("Die Kodierung muss 'SCT' sein");
    }

    /**
     * @return the bic
     */
    public String getBic() {
        return this.bic;
    }

    /**
     * @param bic
     *            the BIC to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn die BIC nicht der Definition entspricht
     */
    public void setBic(final String bic) throws BcdDataTransferObjcetDefinitionException {
        if (bic == null) {
            throw new BcdDataTransferObjcetDefinitionException("Die BIC darf nich NULL sein.");
        }
        if (bic.length() == 0 && this.version.equals("002")) {
            this.bic = bic;
        } else {
            if (IbanAndBicValidator.isBIC(bic.toUpperCase(Locale.getDefault())) == IbanAndBicValidator.NOT_A_BIC) {
                throw new BcdDataTransferObjcetDefinitionException("Die BIC hat ein falsches Format.");
            }
            this.bic = bic.toUpperCase(Locale.getDefault());
        }
    }

    /**
     * @return the empfaenger
     */
    public String getEmpfaenger() {
        return this.empfaenger;
    }

    /**
     * @param empfaenger
     *            the empfaenger to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn der Empfnger mehr als 70 Zeichen lang ist.
     */
    public void setEmpfaenger(final String empfaenger) throws BcdDataTransferObjcetDefinitionException {
        if (empfaenger == null) {
            throw new BcdDataTransferObjcetDefinitionException("Der Empfnger darf nicht null sein.");
        }
        if (empfaenger.length() > 70) { // NOPMD
            throw new BcdDataTransferObjcetDefinitionException(
                    "Der Empfnger darf nicht mehr als 70 Zeichen lang sein.");
        }
        this.empfaenger = empfaenger;
    }

    /**
     * @return the iban
     */
    public String getIban() {
        return this.iban;
    }

    /**
     * @param iban
     *            the iban to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn die IBAN nicht der Definition entspricht
     */
    public void setIban(final String iban) throws BcdDataTransferObjcetDefinitionException {
        if (iban == null) {
            throw new BcdDataTransferObjcetDefinitionException("Die IBAN darf nicht null sein.");
        }
        final int result = IbanAndBicValidator.isIBAN(iban.toUpperCase(Locale.getDefault()));
        if (result == IbanAndBicValidator.NOT_AN_IBAN) {
            throw new BcdDataTransferObjcetDefinitionException("IBAN hat kein gltiges Format.");
        }
        if (result == IbanAndBicValidator.BAD_IBAN_CHECKNUM) {
            throw new BcdDataTransferObjcetDefinitionException("Checksumme ser IBAN ist ungltig.");
        }
        this.iban = iban.toUpperCase(Locale.getDefault());
    }

    /**
     * @return the betrag
     */
    public String getBetrag() {
        return this.betrag;
    }

    /**
     * @param betrag
     *            the betrag to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn der Betrag nicht der Definition entspricht
     */
    public void setBetrag(final String betrag) throws BcdDataTransferObjcetDefinitionException {
        String betragString = betrag;
        if (betragString == null) {
            betragString = StringUtils.EMPTY;
        }
        if (!betragString.isEmpty()) {
            if (betragString.toUpperCase(Locale.getDefault()).startsWith("EUR")) {
                betragString = betragString.substring(3);
            }
            try {
                final Pattern pattern = Pattern.compile("\\d{1,9}(.\\d{1,2})?$");
                final Matcher matcher = pattern.matcher(betragString);
                final boolean isNumber = matcher.find();
                if (!isNumber) {
                    throw new BcdDataTransferObjcetDefinitionException(
                            "Betrag-Definition stimmt nicht mit dem erlaubtem Format EUR999999999.99 berein.");
                }
            } catch (final PatternSyntaxException e) {
                throw new BcdDataTransferObjcetDefinitionException("Betrag-Definition stimmt nicht", e);
            }
        }
        this.betrag = betragString;
    }

    /**
     * @return the geschaeftsCode
     */
    public String getGeschaeftsCode() {
        return this.geschaeftsCode;
    }

    /**
     * @param geschaeftsCode
     *            the geschaeftsCode to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn der Geschftscode nicht der Definition entspricht
     */
    public void setGeschaeftsCode(final String geschaeftsCode) throws BcdDataTransferObjcetDefinitionException {
        String code = geschaeftsCode;
        if (code == null) {
            code = StringUtils.EMPTY;
        }
        if (!code.isEmpty() && code.length() != 4) {
            throw new BcdDataTransferObjcetDefinitionException("Geschaeftscode muss 4 Zeichen lang sein.");
        }
        this.geschaeftsCode = code;
    }

    /**
     * @return the referenze
     */
    public String getReferenze() {
        return this.referenze;
    }

    /**
     * @param referenze
     *            the Referenz to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn die Referenz zu lang (mehr als 35 Zeichen) oder der Verwendungszweck bereits
     *             gesetzt ist
     */
    public void setReferenze(final String referenze) throws BcdDataTransferObjcetDefinitionException {
        String ref = referenze;
        if (ref == null) {
            ref = "";
        }
        if (ref.length() > 35) { // NOPMD
            throw new BcdDataTransferObjcetDefinitionException(
                    "Die Referenze darf nicht mehr als 35 Zeichen lang sein. ");
        }
        // MANTISBT-0000001: Erlaubt nun mit Warnung
        // if (!referenze.isEmpty() && this.verwendungszweck.length() > 0)
        // throw new BcdDataTransferObjcetDefinitionException(
        // "Die Referenz kann nur gesetzt werden, wenn der Verwendungszweck leer ist.");
        this.referenze = ref;
    }

    /**
     * @return the verwendungszweck
     */
    public String getVerwendungszweck() {
        return this.verwendungszweck;
    }

    /**
     * @param verwendungszweck
     *            the verwendungszweck to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn der Verwendungszweck zu lang (mehr als 140 Zeichen) oder die Referenz bereits
     *             gesetzt ist
     */
    public void setVerwendungszweck(final String verwendungszweck) throws BcdDataTransferObjcetDefinitionException {
        String zweck = verwendungszweck;
        if (zweck == null) {
            zweck = StringUtils.EMPTY;
        }
        if (zweck.length() > BcdDataTransferObject.ZWECK_MAX_LENGTH) {
            throw new BcdDataTransferObjcetDefinitionException(
                    "Der Verwendungszweck darf nicht mehr als 140 Zeichen lang sein.");
        }
        // MANTISBT-0000001. Erlaubt nun mit Warnung
        // if (!verwendungszweck.isEmpty() && this.referenze.length() > 0)
        // throw new BcdDataTransferObjcetDefinitionException(
        // "Der Verwendungszweck kann nur gesetzt werden, wenn die Referenze leer ist.");
        this.verwendungszweck = zweck;
    }

    /**
     * @return the benutzerhinweis
     */
    public String getBenutzerhinweis() {
        return this.benutzerhinweis;
    }

    /**
     * @param benutzerhinweis
     *            the benutzerhinweis to set
     * @throws BcdDataTransferObjcetDefinitionException
     *             wird erzeugt, wenn der Benutzerhnweis mehr als 70 Zeichen lang ist
     */
    public void setBenutzerhinweis(final String benutzerhinweis) throws BcdDataTransferObjcetDefinitionException {
        String hinweis = benutzerhinweis;
        if (hinweis == null) {
            hinweis = StringUtils.EMPTY;
        }
        if (hinweis.length() > BcdDataTransferObject.HINT_MAX_LENGTH) {
            throw new BcdDataTransferObjcetDefinitionException(
                    "Der Benutzhinweis darf nicht mehr als 70 Zeichen lang sein.");
        }
        this.benutzerhinweis = hinweis;
    }

    /**
     * Erstellt eine neue {@link de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung} und ffent diese in einer
     * {@link de.willuhn.jameica.hbci.gui.views.AuslandsUeberweisungNew}
     */
    public void oeffneNeueSEPA() {
        try {
            final AuslandsUeberweisung sepa = this.getSEPA();
            if (sepa != null) {
                // View neu zeichnen
                GUI.startView(de.willuhn.jameica.hbci.gui.views.AuslandsUeberweisungNew.class, sepa);
                if (this.getBenutzerhinweis().length() > 0) {
                    // zeige Benutzerhinweis
                    final String text = "Hinweise des QrCode-Erstellers:\n\n" + this.getBenutzerhinweis();
                    Application.getCallback().notifyUser(text);
                }
            }
        } catch (final RemoteException | NumberFormatException e) {
            Logger.error("Die SEPA-berweisung konnte nicht erstellt werden. " + e.getClass().getSimpleName() + ": "
                    + e.getMessage());
        } catch (final Exception e) { // NOPMD
            Logger.error("Info-Fenster konnte nicht geffnet werden. Fehler:\n" + e.getMessage());
        }
    }

    /**
     * Liefert eine SEPA-berweisung mit den Daten des Transfer-Objekts
     *
     * @return die neue SEPA-berweisung
     * @throws RemoteException
     */
    private AuslandsUeberweisung getSEPA() throws RemoteException {
        final AuslandsUeberweisung sepa = DefaultSepaTransaction.getDefaultSepaTransaction();
        sepa.setGegenkontoName(this.getEmpfaenger());
        sepa.setGegenkontoNummer(this.getIban());
        sepa.setGegenkontoBLZ(this.getBic());
        if (StringUtils.isEmpty(this.geschaeftsCode)) {
            this.geschaeftsCode = Settings.getPursposeCode();
        }
        sepa.setPurposeCode(this.getGeschaeftsCode());
        if (!this.getBetrag().isEmpty()) {
            sepa.setBetrag(Double.valueOf(this.getBetrag().replace(",", ".")));
        }
        // MANTISBT-0000001: Behandlung der Kunden-Referenz bzw. End-to-End ID
        if (this.getReferenze().length() > 0 && this.getVerwendungszweck().length() > 0) {
            // zeige Benutzerhinweis
            final String text = "Der QR-Code ist fehlerhaft!!!\n\n"
                    + "Er enthlt sowohl eine Refernzangebe als auch einen Verwendungszweck.\n"
                    + "Dies ist laut Spezifikation nicht zulssig. Bitte informieren sie den\n"
                    + "Rechungsersteller.\n\n"
                    + "Die Informationen knnen zusammengefasst als Verwendungszweck bernommen\n"
                    + "werden. Da dies zu einer ungltigen berweisung fhren kann, berprfen\n"
                    + "sie  die bernommen Daten bitte genau.\n\n\n" + "Daten als Verwendungszweck bernehmen?\n\n ";
            boolean takeboth = true;
            try {
                takeboth = Application.getCallback().askUser(text, false);
            } catch (final Exception e) { // NOPMD
                Logger.error("Benutzerabfrage fehlgeschlagen. Es werden beide Text verwendet", e);
            }
            if (takeboth) {
                sepa.setZweck(this.getReferenze() + " - " + this.getVerwendungszweck());
            }
        } else if (this.getReferenze().length() > 0) {
            sepa.setZweck(this.getReferenze());
            sepa.setEndtoEndId(this.getReferenze());
            sepa.setPmtInfId(this.getReferenze());
        } else {
            sepa.setZweck(this.getVerwendungszweck());
        }
        return sepa;
    }
}
